@@ -1,19 +1,23 @@
 import enum
 import logging
+import re
 from datetime import timedelta
 from typing import Dict, Optional, Union
 from dataclasses import dataclass
 
 import yaml
+from asn1crypto import x509
+
 from certvalidator import ValidationContext
 from pyhanko.pdf_utils.config_utils import (
     check_config_keys, ConfigurationError
 )
 from pyhanko.pdf_utils.misc import get_and_apply
-from pyhanko.sign import signers
 
 
 # TODO add stamp styles etc.
+from pyhanko.sign import load_certs_from_pemder
+from pyhanko.sign.general import KeyUsageConstraints
 from pyhanko.sign.signers import DEFAULT_SIGNING_STAMP_STYLE
 from pyhanko.stamp import QRStampStyle, TextStampStyle
 
@@ -57,21 +61,50 @@ class CLIConfig:
     default_validation_context: str
     default_stamp_style: str
     time_tolerance: timedelta
+    retroactive_revinfo: bool
     log_config: Dict[Optional[str], LogConfig]
 
-    # TODO graceful error handling
+    # TODO graceful error handling for syntax & type issues?
 
-    def get_validation_context(self, name=None, as_dict=False):
+    def _get_validation_settings_raw(self, name=None):
         name = name or self.default_validation_context
         try:
-            vc_config = self.validation_contexts[name]
+            return self.validation_contexts[name]
         except KeyError:
             raise ConfigurationError(
                 f"There is no validation context named '{name}'."
             )
-        vc_kwargs = parse_trust_config(vc_config, self.time_tolerance)
+
+    def get_validation_context(self, name=None, as_dict=False):
+        vc_config = self._get_validation_settings_raw(name)
+        vc_kwargs = parse_trust_config(
+            vc_config, self.time_tolerance, self.retroactive_revinfo
+        )
         return vc_kwargs if as_dict else ValidationContext(**vc_kwargs)
 
+    def get_signer_key_usages(self, name=None) -> KeyUsageConstraints:
+        vc_config = self._get_validation_settings_raw(name)
+
+        try:
+            key_usage_strings = vc_config['signer-key-usage']
+            key_usage = set(
+                _process_key_usages(key_usage_strings, extended=False)
+            )
+        except KeyError:
+            key_usage = None
+
+        try:
+            key_usage_strings = vc_config['signer-extd-key-usage']
+            extd_key_usage = set(
+                _process_key_usages(key_usage_strings, extended=True)
+            )
+        except KeyError:
+            extd_key_usage = None
+
+        return KeyUsageConstraints(
+            key_usage=key_usage, extd_key_usage=extd_key_usage
+        )
+
     def get_stamp_style(self, name=None) -> TextStampStyle:
         name = name or self.default_stamp_style
         try:
@@ -86,18 +119,26 @@ class CLIConfig:
         return cls.from_config(style_config)
 
 
-def init_validation_context_kwargs(trust, trust_replace, other_certs,
+def init_validation_context_kwargs(*, trust, trust_replace, other_certs,
+                                   retroactive_revinfo=False,
                                    time_tolerance=None):
-    vc_kwargs = {
-        'time_tolerance':
-            timedelta(seconds=DEFAULT_TIME_TOLERANCE) if time_tolerance is None
-            else time_tolerance
-    }
+    if not isinstance(time_tolerance, timedelta):
+        if time_tolerance is None:
+            time_tolerance = timedelta(seconds=DEFAULT_TIME_TOLERANCE)
+        elif isinstance(time_tolerance, int):
+            time_tolerance = timedelta(seconds=time_tolerance)
+        else:
+            raise ConfigurationError(
+                "time-tolerance parameter must be specified in seconds"
+            )
+    vc_kwargs = {'time_tolerance': time_tolerance}
+    if retroactive_revinfo:
+        vc_kwargs['retroactive_revinfo'] = True
     if trust:
         if isinstance(trust, str):
             trust = (trust,)
         # add trust roots to the validation context, or replace them
-        trust_certs = list(signers.load_certs_from_pemder(trust))
+        trust_certs = list(load_certs_from_pemder(trust))
         if trust_replace:
             vc_kwargs['trust_roots'] = trust_certs
         else:
@@ -105,9 +146,7 @@ def init_validation_context_kwargs(trust, trust_replace, other_certs,
     if other_certs:
         if isinstance(other_certs, str):
             other_certs = (other_certs,)
-        vc_kwargs['other_certs'] = list(
-            signers.load_certs_from_pemder(other_certs)
-        )
+        vc_kwargs['other_certs'] = list(load_certs_from_pemder(other_certs))
     return vc_kwargs
 
 
@@ -115,19 +154,67 @@ def init_validation_context_kwargs(trust, trust_replace, other_certs,
 #  in some cases)
 # Time-related settings are probably better off in the CLI.
 
-def parse_trust_config(trust_config, time_tolerance) -> dict:
+def parse_trust_config(trust_config, time_tolerance,
+                       retroactive_revinfo) -> dict:
     check_config_keys(
-        'ValidationContext', ('trust', 'trust-replace', 'other-certs'),
+        'ValidationContext',
+        ('trust', 'trust-replace', 'other-certs',
+         'time-tolerance', 'retroactive-revinfo',
+         'signer-key-usage', 'signer-extd-key-usage'),
         trust_config
     )
     return init_validation_context_kwargs(
         trust=trust_config.get('trust'),
         trust_replace=trust_config.get('trust-replace', False),
         other_certs=trust_config.get('other-certs'),
-        time_tolerance=time_tolerance
+        time_tolerance=trust_config.get('time-tolerance', time_tolerance),
+        retroactive_revinfo=trust_config.get(
+            'retroactive-revinfo', retroactive_revinfo
+        )
     )
 
 
+OID_REGEX = re.compile(r'\d(\.\d+)+')
+
+
+def _process_key_usages(key_usage_strings, extended=False):
+    err_msg = (
+        "Key usages and extended key usages must be specified as a "
+        "list of strings, or a string."
+    )
+    if isinstance(key_usage_strings, str):
+        key_usage_strings = (key_usage_strings,)
+    elif not isinstance(key_usage_strings, list):
+        raise ConfigurationError(err_msg)
+
+    usage_class = x509.KeyPurposeId if extended else x509.KeyUsage
+    valid_usages = usage_class._map.values()
+    for usage_string in key_usage_strings:
+        if not isinstance(usage_string, str):
+            raise ConfigurationError(
+                f"{err_msg} '{repr(usage_string)}' is not a string."
+            )
+        if usage_string in valid_usages:
+            yield usage_string
+        elif extended and OID_REGEX.fullmatch(usage_string):
+            # for extended key usages, we allow OIDs as well
+            # and natively translate them to human-readable form
+            # whenever possible
+            try:
+                usage_string = x509.KeyPurposeId._map[usage_string]
+            except KeyError:
+                logging.debug(
+                    f"Extended key usage extension {usage_string} was not "
+                    f"recognised by asn1crypto; using raw value."
+                )
+            yield usage_string
+        else:
+            extd = "extended " if extended else ""
+            raise ConfigurationError(
+                f"'{usage_string}' is not a valid {extd}key usage designation."
+            )
+
+
 DEFAULT_ROOT_LOGGER_LEVEL = logging.INFO
 
 
@@ -221,12 +308,19 @@ def parse_cli_config(yaml_str):
     default_stamp_style = config_dict.get(
         'default-stamp-style', DEFAULT_STAMP_STYLE
     )
-    time_tolerance = timedelta(
-        seconds=config_dict.get('time-tolerance', DEFAULT_TIME_TOLERANCE)
+    time_tolerance_seconds = config_dict.get(
+        'time-tolerance', DEFAULT_TIME_TOLERANCE
     )
+    if not isinstance(time_tolerance_seconds, int):
+        raise ConfigurationError(
+            "time-tolerance parameter must be specified in seconds"
+        )
+
+    time_tolerance = timedelta(seconds=time_tolerance_seconds)
+    retroactive_revinfo = bool(config_dict.get('retroactive-revinfo', False))
     return CLIConfig(
         validation_contexts=vcs, default_validation_context=default_vc,
-        time_tolerance=time_tolerance, stamp_styles=stamp_configs,
-        default_stamp_style=default_stamp_style,
+        time_tolerance=time_tolerance, retroactive_revinfo=retroactive_revinfo,
+        stamp_styles=stamp_configs, default_stamp_style=default_stamp_style,
         log_config=log_config
     )
