@@ -64,7 +64,12 @@ from typing import Dict, Type, Optional, Tuple, Union, List, Set
 
 from asn1crypto import x509, cms, algos
 from asn1crypto.keys import PublicKeyAlgorithm, PrivateKeyInfo
-from oscrypto import symmetric, asymmetric, keys as oskeys
+from cryptography.hazmat.primitives import padding, serialization
+from cryptography.hazmat.primitives.asymmetric.padding import PKCS1v15
+from cryptography.hazmat.primitives.asymmetric.rsa import RSAPublicKey, \
+    RSAPrivateKey
+from cryptography.hazmat.primitives.serialization import pkcs12
+from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
 
 from . import generic, misc
 
@@ -166,6 +171,9 @@ def _legacy_normalise_pw(password: Union[str, bytes]) -> bytes:
 
 def _r6_normalise_pw(password: Union[str, bytes]) -> bytes:
     if isinstance(password, str):
+        # saslprep expects non-empty strings, apparently
+        if not password:
+            return b''
         from ._saslprep import saslprep
         password = saslprep(password).encode('utf-8')
     return password[:127]
@@ -177,12 +185,41 @@ def _r6_password_authenticate(pw_bytes: bytes, entry: _R6KeyEntry,
     return purported_hash == entry.hash_value
 
 
+def _aes_cbc_decrypt(key, data, iv, use_padding=True):
+    cipher = Cipher(algorithms.AES(key), modes.CBC(iv))
+    decryptor = cipher.decryptor()
+    plaintext = decryptor.update(data) + decryptor.finalize()
+
+    if use_padding:
+        unpadder = padding.PKCS7(128).unpadder()
+        return unpadder.update(plaintext) + unpadder.finalize()
+    else:
+        return plaintext
+
+
+def _aes_cbc_encrypt(key, data, iv, use_padding=True):
+    if iv is None:
+        iv = secrets.token_bytes(16)
+    cipher = Cipher(algorithms.AES(key), modes.CBC(iv))
+    encryptor = cipher.encryptor()
+    if use_padding:
+        padder = padding.PKCS7(128).padder()
+        data = padder.update(data) + padder.finalize()
+    return iv, encryptor.update(data) + encryptor.finalize()
+
+
+def _rc4_encrypt(key, data):
+    cipher = Cipher(algorithms.ARC4(key), mode=None)
+    encryptor = cipher.encryptor()
+    return encryptor.update(data) + encryptor.finalize()
+
+
 def _r6_derive_file_key(pw_bytes: bytes, entry: _R6KeyEntry, e_entry: bytes,
                         u_entry: Optional[bytes] = None):
     interm_key = _r6_hash_algo(pw_bytes, entry.key_salt, u_entry)
     assert len(e_entry) == 32
-    return symmetric.aes_cbc_no_padding_decrypt(
-        key=interm_key, data=e_entry, iv=bytes(16)
+    return _aes_cbc_decrypt(
+        key=interm_key, data=e_entry, iv=bytes(16), use_padding=False
     )
 
 
@@ -211,8 +248,8 @@ def _r6_hash_algo(pw_bytes: bytes, current_salt: bytes,
     round_no = last_byte_val = 0
     while round_no < 64 or last_byte_val > round_no - 32:
         k1 = (pw_bytes + k + (u_entry or b'')) * 64
-        e = symmetric.aes_cbc_no_padding_encrypt(
-            key=k[:16], data=k1, iv=k[16:32]
+        e = _aes_cbc_encrypt(
+            key=k[:16], data=k1, iv=k[16:32], use_padding=False
         )[1]
         # compute the first 16 bytes of e, interpreted as an unsigned integer
         # mod 3
@@ -233,7 +270,7 @@ def _compute_o_value_legacy(owner_pwd, user_pwd, rev, keylen):
     user_pwd = (user_pwd + _encryption_padding)[:32]
     # 6. Encrypt the result of step 5, using an RC4 encryption function with
     # the encryption key obtained in step 4.
-    val = symmetric.rc4_encrypt(key, user_pwd)
+    val = _rc4_encrypt(key, user_pwd)
     # 7. (Revision 3 or greater) Do the following 19 times: Take the output
     # from the previous invocation of the RC4 function and pass it as input to
     # a new invocation of the function; use an encryption key generated by
@@ -243,7 +280,7 @@ def _compute_o_value_legacy(owner_pwd, user_pwd, rev, keylen):
     if rev >= 3:
         for i in range(1, 20):
             new_key = bytes(b ^ i for b in key)
-            val = symmetric.rc4_encrypt(new_key, val)
+            val = _rc4_encrypt(new_key, val)
     # 8. Store the output from the final invocation of the RC4 as the value of
     # the /O entry in the encryption dictionary.
     return val
@@ -282,7 +319,7 @@ def _compute_u_value_r2(password, owner_entry, p_entry, id1_entry):
     # 2. Encrypt the 32-byte padding string shown in step 1 of algorithm 3.2,
     # using an RC4 encryption function with the encryption key from the
     # preceding step.
-    u = symmetric.rc4_encrypt(key, _encryption_padding)
+    u = _rc4_encrypt(key, _encryption_padding)
     # 3. Store the result of step 2 as the value of the /U entry in the
     # encryption dictionary.
     return u, key
@@ -308,7 +345,7 @@ def _compute_u_value_r34(password, rev, keylen, owner_entry, p_entry,
     md5_hash = m.digest()
     # 4. Encrypt the 16-byte result of the hash, using an RC4 encryption
     # function with the encryption key from step 1.
-    val = symmetric.rc4_encrypt(key, md5_hash)
+    val = _rc4_encrypt(key, md5_hash)
     # 5. Do the following 19 times: Take the output from the previous
     # invocation of the RC4 function and pass it as input to a new invocation
     # of the function; use an encryption key generated by taking each byte of
@@ -317,7 +354,7 @@ def _compute_u_value_r34(password, rev, keylen, owner_entry, p_entry,
     # counter (from 1 to 19).
     for i in range(1, 20):
         new_key = bytes(b ^ i for b in key)
-        val = symmetric.rc4_encrypt(new_key, val)
+        val = _rc4_encrypt(new_key, val)
     # 6. Append 16 bytes of arbitrary padding to the output from the final
     # invocation of the RC4 function and store the 32-byte result as the value
     # of the U entry in the encryption dictionary.
@@ -841,7 +878,8 @@ class PubKeyCryptFilter(CryptFilter, abc.ABC):
         super()._set_security_handler(handler)
         self._shared_key = self._recp_key_seed = None
 
-    def add_recipients(self, certs: List[x509.Certificate], perms=ALL_PERMS):
+    def add_recipients(self, certs: List[x509.Certificate], perms=ALL_PERMS,
+                       ignore_key_usage=False):
         """
         Add recipients to this crypt filter.
         This always adds one full CMS object to the Recipients array
@@ -850,6 +888,8 @@ class PubKeyCryptFilter(CryptFilter, abc.ABC):
             A list of recipient certificates.
         :param perms:
             The permission bits to assign to the listed recipients.
+        :param ignore_key_usage:
+            If ``False``, the *keyEncipherment* key usage extension is required.
         """
 
         if not self.acts_as_default and self.recipients:
@@ -870,7 +910,8 @@ class PubKeyCryptFilter(CryptFilter, abc.ABC):
             )
         new_cms = construct_recipient_cms(
             certs, self._recp_key_seed, _as_signed(perms),
-            include_permissions=self.acts_as_default
+            include_permissions=self.acts_as_default,
+            ignore_key_usage=ignore_key_usage
         )
         self.recipients.append(new_cms)
 
@@ -1032,7 +1073,7 @@ class RC4CryptFilterMixin(CryptFilter, abc.ABC):
         :return:
             Ciphertext.
         """
-        return symmetric.rc4_encrypt(key, plaintext)
+        return _rc4_encrypt(key, plaintext)
 
     def decrypt(self, key, ciphertext: bytes, params=None) -> bytes:
         """
@@ -1047,7 +1088,7 @@ class RC4CryptFilterMixin(CryptFilter, abc.ABC):
         :return:
             Plaintext.
         """
-        return symmetric.rc4_encrypt(key, ciphertext)
+        return _rc4_encrypt(key, ciphertext)
 
     def derive_object_key(self, idnum, generation) -> bytes:
         """
@@ -1093,7 +1134,7 @@ class AESCryptFilterMixin(CryptFilter, abc.ABC):
             The resulting ciphertext, prepended with a 16-byte initialisation
             vector.
         """
-        iv, ciphertext = symmetric.aes_cbc_pkcs7_encrypt(
+        iv, ciphertext = _aes_cbc_encrypt(
             key, plaintext, secrets.token_bytes(16)
         )
         return iv + ciphertext
@@ -1113,7 +1154,7 @@ class AESCryptFilterMixin(CryptFilter, abc.ABC):
             The resulting plaintext.
         """
         iv, data = ciphertext[:16], ciphertext[16:]
-        return symmetric.aes_cbc_pkcs7_decrypt(key, data, iv)
+        return _aes_cbc_decrypt(key, data, iv)
 
     def derive_object_key(self, idnum, generation) -> bytes:
         """
@@ -1483,8 +1524,8 @@ class StandardSecurityHandler(SecurityHandler):
         u_hash = _r6_hash_algo(user_pw_bytes, u_validation_salt)
         u_entry = u_hash + u_validation_salt + u_key_salt
         u_interm_key = _r6_hash_algo(user_pw_bytes, u_key_salt)
-        _, ue_seed = symmetric.aes_cbc_no_padding_encrypt(
-            u_interm_key, encryption_key, bytes(16)
+        _, ue_seed = _aes_cbc_encrypt(
+            u_interm_key, encryption_key, bytes(16), use_padding=False
         )
         assert len(ue_seed) == 32
 
@@ -1493,8 +1534,8 @@ class StandardSecurityHandler(SecurityHandler):
         o_hash = _r6_hash_algo(owner_pw_bytes, o_validation_salt, u_entry)
         o_entry = o_hash + o_validation_salt + o_key_salt
         o_interm_key = _r6_hash_algo(owner_pw_bytes, o_key_salt, u_entry)
-        _, oe_seed = symmetric.aes_cbc_no_padding_encrypt(
-            o_interm_key, encryption_key, bytes(16)
+        _, oe_seed = _aes_cbc_encrypt(
+            o_interm_key, encryption_key, bytes(16), use_padding=False
         )
         assert len(oe_seed) == 32
 
@@ -1503,23 +1544,15 @@ class StandardSecurityHandler(SecurityHandler):
             perms_bytes + (b'\xff' * 4) + b'Tadb' + secrets.token_bytes(4)
         )
 
-        # need to encrypt one 16 byte block in CBC mode with an
-        # IV of 0 (equivalent to 1 block in ECB mode).
-        # FIXME
-        #  oscrypto's OpenSSL-based implementation of aes_cbc_no_padding_encrypt
-        #  requires padding to encrypt a single 16-byte block if the key size is
-        #  larger than 16 bytes. I believe that's a mistake, but since I'm not
-        #  willing to vendor oscrypto, this little hack will have to do.
+        # need to encrypt one 16 byte block in ECB mode
+        #  [I _really_ don't like the way this part of the spec works, but
+        #   we have to sacrifice our principles on the altar of backwards
+        #   compatibility.]
+        cipher = Cipher(algorithms.AES(encryption_key), modes.ECB())
+        encryptor = cipher.encryptor()
+        encrypted_perms = \
+            encryptor.update(extd_perms_bytes) + encryptor.finalize()
 
-        # We'll indulge in oscrypto's whims and request padding
-        _, encrypted_perms = symmetric.aes_cbc_pkcs7_encrypt(
-            encryption_key, extd_perms_bytes, bytes(16)
-        )
-
-        # ... and then cut the result off at 16 bytes
-        # This is OK, because the block size is 16 bytes, so the first 16
-        # bytes aren't affected by the rest of the encrypted string
-        encrypted_perms = encrypted_perms[:16]
         sh = StandardSecurityHandler(
             version=SecurityHandlerVersion.AES256,
             revision=StandardSecuritySettingsRevision.AES256,
@@ -1707,12 +1740,12 @@ class StandardSecurityHandler(SecurityHandler):
             rev = self.revision
             key = _compute_o_value_legacy_prep(password, rev.value, self.keylen)
             if rev == StandardSecuritySettingsRevision.RC4_BASIC:
-                userpass = symmetric.rc4_encrypt(key, self.odata)
+                userpass = _rc4_encrypt(key, self.odata)
             else:
                 val = self.odata
                 for i in range(19, -1, -1):
                     new_key = bytes(b ^ i for b in key)
-                    val = symmetric.rc4_encrypt(new_key, val)
+                    val = _rc4_encrypt(new_key, val)
                 userpass = val
             owner_password, key = self._auth_user_password_legacy(id1, userpass)
             if owner_password:
@@ -1769,13 +1802,14 @@ class StandardSecurityHandler(SecurityHandler):
         else:
             return AuthStatus.FAILED, None
 
-        # check the file key against the perms entry
-
-        # Standard says ECB (which oscrypto doesn't support),
-        # but one round of CBC with IV = 0 is equivalent to ECB
-        decrypted_p_entry = symmetric.aes_cbc_no_padding_decrypt(
-            key, self.encrypted_perms, bytes(16)
-        )
+        # need to encrypt one 16 byte block in ECB mode
+        #  [I _really_ don't like the way this part of the spec works, but
+        #   we have to sacrifice our principles on the altar of backwards
+        #   compatibility.]
+        cipher = Cipher(algorithms.AES(key), modes.ECB())
+        decryptor = cipher.decryptor()
+        decrypted_p_entry = \
+            decryptor.update(self.encrypted_perms) + decryptor.finalize()
 
         # known plaintext mandated in the standard ...sigh...
         perms_ok = decrypted_p_entry[9:12] == b'adb'
@@ -1830,7 +1864,8 @@ def construct_envelope_content(seed: bytes, perms: int,
     return seed + (struct.pack('<i', perms) if include_permissions else b'')
 
 
-def _recipient_info(envelope_key: bytes, cert: x509.Certificate):
+def _recipient_info(envelope_key: bytes, cert: x509.Certificate,
+                    ignore_key_usage=False):
     pubkey = cert.public_key
     pubkey_algo_info: PublicKeyAlgorithm = pubkey['algorithm']
     algorithm_name = pubkey_algo_info['algorithm'].native
@@ -1842,11 +1877,24 @@ def _recipient_info(envelope_key: bytes, cert: x509.Certificate):
 
     assert len(envelope_key) == 32
 
-    # TODO having support for oeap here would be cool, but as with PSS
-    #  oscrypto only supports the default parameters.
-    encrypted_data = asymmetric.rsa_pkcs1v15_encrypt(
-        asymmetric.load_public_key(pubkey), envelope_key
+    if not ignore_key_usage:
+        key_usage = cert.key_usage_value
+        if key_usage is None or 'key_encipherment' not in key_usage.native:
+            raise misc.PdfWriteError(
+                f"Certificate for subject {cert.subject.human_friendly} does "
+                f"not have the 'key_encipherment' key usage bit set."
+            )
+
+    pub_key = serialization.load_der_public_key(
+        cert.public_key.dump()
     )
+
+    assert isinstance(pub_key, RSAPublicKey)
+    # having support for OAEP here would be cool, but I have it on good
+    #  authority that there's some kind of tacit understanding to use
+    #  PKCS#1 v1.5 padding here.
+    encrypted_data = pub_key.encrypt(envelope_key, padding=PKCS1v15())
+
     # TODO support subjectKeyIdentifier here (requiring version 2)
     rid = cms.RecipientIdentifier({
         'issuer_and_serial_number': cms.IssuerAndSerialNumber({
@@ -1865,7 +1913,8 @@ def _recipient_info(envelope_key: bytes, cert: x509.Certificate):
 
 
 def construct_recipient_cms(certificates: List[x509.Certificate], seed: bytes,
-                            perms: int, include_permissions=True) \
+                            perms: int, include_permissions=True,
+                            ignore_key_usage=False) \
         -> cms.ContentInfo:
 
     # The content of the generated ContentInfo object
@@ -1889,12 +1938,15 @@ def construct_recipient_cms(certificates: List[x509.Certificate], seed: bytes,
     # 256-bit key used to encrypt the envelope
     envelope_key = secrets.token_bytes(32)
     # encrypt the envelope content with the envelope key
-    iv, encrypted_envelope_content = symmetric.aes_cbc_pkcs7_encrypt(
+    iv, encrypted_envelope_content = _aes_cbc_encrypt(
         envelope_key, envelope_content, iv=None
     )
 
     # encrypt the envelope key for each recipient
-    rec_infos = [_recipient_info(envelope_key, cert) for cert in certificates]
+    rec_infos = [
+        _recipient_info(envelope_key, cert, ignore_key_usage=ignore_key_usage)
+        for cert in certificates
+    ]
 
     algo = cms.EncryptionAlgorithm({
         'algorithm': algos.EncryptionAlgorithmId('aes256_cbc'),
@@ -1958,12 +2010,12 @@ class SimpleEnvelopeKeyDecrypter(EnvelopeKeyDecrypter):
     :param cert:
         The recipient's certificate.
     :param private_key:
-        The recipient's private key, as a CMS object.
+        The recipient's private key.
     """
 
     def __init__(self, cert: x509.Certificate, private_key: PrivateKeyInfo):
         super().__init__(cert)
-        self.private_key = private_key
+        self.private_key: PrivateKeyInfo = private_key
 
     @staticmethod
     def load(key_file, cert_file, key_passphrase=None):
@@ -1979,17 +2031,15 @@ class SimpleEnvelopeKeyDecrypter(EnvelopeKeyDecrypter):
         :return:
             An instance of :class:`.SimpleEnvelopeKeyDecrypter`.
         """
+        from pyhanko.sign.general import load_private_key_from_pemder
+
         try:
-            # load cryptographic data (both PEM and DER are supported)
-            with open(key_file, 'rb') as f:
-                private_key: PrivateKeyInfo = oskeys.parse_private(
-                    f.read(), password=key_passphrase
-                )
-            with open(cert_file, 'rb') as f:
-                cert: x509.Certificate = oskeys.parse_certificate(
-                    f.read()
-                )
-        except (IOError, ValueError) as e:  # pragma: nocover
+            private_key = load_private_key_from_pemder(
+                key_file, passphrase=key_passphrase
+            )
+            from pyhanko.sign.general import load_cert_from_pemder
+            cert = load_cert_from_pemder(cert_file)
+        except (IOError, ValueError, TypeError) as e:  # pragma: nocover
             logger.error('Could not load cryptographic material', exc_info=e)
             return None
         return SimpleEnvelopeKeyDecrypter(cert=cert, private_key=private_key)
@@ -2010,13 +2060,21 @@ class SimpleEnvelopeKeyDecrypter(EnvelopeKeyDecrypter):
         try:
             with open(pfx_file, 'rb') as f:
                 pfx_bytes = f.read()
-        except IOError as e:  # pragma: nocover
+            (private_key, cert, other_certs) = pkcs12.load_key_and_certificates(
+                pfx_bytes, passphrase
+            )
+
+            from ..sign.general import (
+                _translate_pyca_cryptography_key_to_asn1,
+                _translate_pyca_cryptography_cert_to_asn1
+            )
+            cert = _translate_pyca_cryptography_cert_to_asn1(cert)
+            private_key = _translate_pyca_cryptography_key_to_asn1(private_key)
+        except (IOError, ValueError, TypeError) as e:  # pragma: nocover
             logger.error(f'Could not open PKCS#12 file {pfx_file}.', exc_info=e)
             return None
 
-        (kinfo, cert, other_certs) = oskeys.parse_pkcs12(pfx_bytes, passphrase)
-
-        return SimpleEnvelopeKeyDecrypter(cert=cert, private_key=kinfo)
+        return SimpleEnvelopeKeyDecrypter(cert=cert, private_key=private_key)
 
     def decrypt(self, encrypted_key: bytes,
                 algo_params: cms.KeyEncryptionAlgorithm) -> bytes:
@@ -2038,9 +2096,10 @@ class SimpleEnvelopeKeyDecrypter(EnvelopeKeyDecrypter):
                 f"Only 'rsaes_pkcs1v15' is supported for envelope encryption, "
                 f"not '{algo_name}'."
             )
-        return asymmetric.rsa_pkcs1v15_decrypt(
-            asymmetric.load_private_key(self.private_key), encrypted_key
+        priv_key: RSAPrivateKey = serialization.load_der_private_key(
+            self.private_key.dump(), password=None
         )
+        return priv_key.decrypt(encrypted_key, padding=PKCS1v15())
 
 
 def read_seed_from_recipient_cms(recipient_cms: cms.ContentInfo,
@@ -2093,19 +2152,32 @@ def read_seed_from_recipient_cms(recipient_cms: cms.ContentInfo,
     # des, triple des, rc2 (<=128 bits)
     # and AES-CBC (128, 192, 256 bits)
     cipher_name = algo.encryption_cipher
-    with_iv = {
-        'aes': symmetric.aes_cbc_pkcs7_decrypt,
-        'des': symmetric.des_cbc_pkcs5_decrypt,
-        'tripledes': symmetric.tripledes_cbc_pkcs5_decrypt,
-        'rc2': symmetric.rc2_cbc_pkcs5_decrypt
-    }
+
+    with_iv = {'aes': _aes_cbc_decrypt}
+    try:
+        # noinspection PyUnresolvedReferences
+        from oscrypto import symmetric
+        # The spec mandates that we support these, but pyca/cryptography
+        # doesn't offer implementations.
+        # (DES and 3DES have fortunately gone out of style, but some libraries
+        #  still rely on RC2)
+        with_iv.update({
+            'des': symmetric.des_cbc_pkcs5_decrypt,
+            'tripledes': symmetric.tripledes_cbc_pkcs5_decrypt,
+            'rc2': symmetric.rc2_cbc_pkcs5_decrypt
+        })
+    except ImportError:
+        if cipher_name in ('des', 'tripledes', 'rc2'):
+            raise NotImplementedError(
+                "DES, 3DES and RC2 require oscrypto to be present"
+            )
 
     if cipher_name in with_iv:
         decryption_fun = with_iv[cipher_name]
         iv = algo.encryption_iv
         content = decryption_fun(envelope_key, encrypted_envelope_content, iv)
     elif cipher_name == 'rc4':
-        content = symmetric.rc4_encrypt(
+        content = _rc4_encrypt(
             envelope_key, encrypted_envelope_content
         )
     else:
@@ -2136,7 +2208,8 @@ class PubKeySecurityHandler(SecurityHandler):
                          version=SecurityHandlerVersion.AES256,
                          use_aes=True, use_crypt_filters=True,
                          perms: int = ALL_PERMS,
-                         encrypt_metadata=True) -> 'PubKeySecurityHandler':
+                         encrypt_metadata=True,
+                         ignore_key_usage=False) -> 'PubKeySecurityHandler':
         """
         Create a new public key security handler.
 
@@ -2166,6 +2239,8 @@ class PubKeySecurityHandler(SecurityHandler):
             .. warning::
                 See :class:`.SecurityHandlers` for some background on the
                 way pyHanko interprets this value.
+        :param ignore_key_usage:
+            If ``False``, the *keyEncipherment* key usage extension is required.
         :return:
             An instance of :class:`.PubKeySecurityHandler`.
         """
@@ -2192,7 +2267,7 @@ class PubKeySecurityHandler(SecurityHandler):
             encrypt_metadata=encrypt_metadata, crypt_filter_config=cfc,
             recipient_objs=None
         )
-        sh.add_recipients(certs, perms=perms)
+        sh.add_recipients(certs, perms=perms, ignore_key_usage=ignore_key_usage)
         return sh
 
     def __init__(self, version: SecurityHandlerVersion,
@@ -2329,7 +2404,8 @@ class PubKeySecurityHandler(SecurityHandler):
             )
         return result
 
-    def add_recipients(self, certs: List[x509.Certificate], perms=ALL_PERMS):
+    def add_recipients(self, certs: List[x509.Certificate], perms=ALL_PERMS,
+                       ignore_key_usage=False):
         # add recipients to all *default* crypt filters
         # callers that want to do this more granularly are welcome to, but
         # then they have to do the legwork themselves.
@@ -2337,7 +2413,9 @@ class PubKeySecurityHandler(SecurityHandler):
         for cf in self.crypt_filter_config.default_filters():
             if not isinstance(cf, PubKeyCryptFilter):
                 continue
-            cf.add_recipients(certs, perms=perms)
+            cf.add_recipients(
+                certs, perms=perms, ignore_key_usage=ignore_key_usage
+            )
 
     def authenticate(self, credential: EnvelopeKeyDecrypter, id1=None) \
             -> AuthResult:
