@@ -61,7 +61,7 @@ from enum import unique
 from io import BytesIO
 from typing import (
     Iterable, Optional, Set, Tuple, Generator, TypeVar, Dict,
-    List, Callable, Union
+    List, Callable, Union, Iterator
 )
 
 from pyhanko.pdf_utils.generic import Reference, PdfObject
@@ -730,6 +730,8 @@ class SigFieldCreationRule(FieldMDPRule):
         # finally, deal with the signature fields themselves
         # The distinction between timestamps and signatures isn't relevant
         # yet, that's a problem for /V, which we don't bother with here.
+        field_ref_reverse = {}
+
         for fq_name, sigfield_ref in all_new_refs.items():
 
             # new field, so all its dependencies are good to go
@@ -748,9 +750,9 @@ class SigFieldCreationRule(FieldMDPRule):
             # checked by field listing routine already
             assert isinstance(sigfield, generic.DictionaryObject)
 
-            for _key in ('/AP', '/Lock', '/SV'):
+            def _handle_deps(pdf_dict, _key):
                 try:
-                    raw_value = sigfield.raw_get(_key)
+                    raw_value = pdf_dict.raw_get(_key)
 
                     deps = context.new.collect_dependencies(
                         raw_value,
@@ -764,7 +766,42 @@ class SigFieldCreationRule(FieldMDPRule):
                 except KeyError:
                     pass
 
-        # Next, check (widget) annotations: newly added signature fields may
+            for _key in ('/AP', '/Lock', '/SV'):
+                yield from _handle_deps(sigfield, _key)
+
+            # if the field has widget annotations in /Kids, add them to the
+            #  field_ref_reverse dictionary for annotation processing later
+            try:
+                kids_arr_ref = sigfield.raw_get('/Kids')
+                old = context.old
+                if isinstance(kids_arr_ref, generic.IndirectObject) \
+                        and old.is_ref_available(kids_arr_ref.reference):
+                    yield ModificationLevel.LTA_UPDATES, FormUpdate(
+                        updated_ref=kids_arr_ref.reference, field_name=fq_name,
+                        valid_when_locked=True
+                    )
+                kid_refs = _arr_to_refs(
+                    kids_arr_ref.get_object(), SuspiciousModification
+                )
+                # process all widgets in /Kids
+                # in principle there should be only one, but we don't enforce
+                # that restriction here
+                # TODO make that togglable?
+                for kid in kid_refs:
+                    if '/T' not in kid.get_object():
+                        field_ref_reverse[kid] = fq_name
+                        if old.is_ref_available(kid):
+                            yield ModificationLevel.LTA_UPDATES, FormUpdate(
+                                updated_ref=kid, field_name=fq_name,
+                                valid_when_locked=True
+                            )
+                        # pull in appearance dependencies
+                        yield from _handle_deps(kid.get_object(), '/AP')
+            except KeyError:
+                # No /Kids => assume the field is its own annotation
+                field_ref_reverse[sigfield_ref] = fq_name
+
+        # Now we process (widget) annotations: newly added signature fields may
         #  be added to the /Annots entry of any page. These are processed as LTA
         #  updates, because even invisible signature fields / timestamps might
         #  be added to /Annots (this isn't strictly necessary, but more
@@ -786,8 +823,6 @@ class SigFieldCreationRule(FieldMDPRule):
         old_page_root = context.old.root['/Pages']
         new_page_root = context.new.root['/Pages']
 
-        field_ref_reverse = {v: k for k, v in all_new_refs.items()}
-
         yield from qualify(
             ModificationLevel.LTA_UPDATES,
             _walk_page_tree_annots(
@@ -1382,7 +1417,10 @@ def _list_fields(old_fields: generic.PdfObject, new_fields: generic.PdfObject,
             if not isinstance(field, generic.DictionaryObject):
                 raise exc("Fields must be dictionary objects")
 
-            name = field.raw_get('/T')
+            try:
+                name = field.raw_get('/T')
+            except KeyError:
+                continue
             if name in names_seen:
                 raise exc("Duplicate field name")
             elif '.' in name:
@@ -1820,6 +1858,58 @@ class DiffPolicy:
         raise NotImplementedError
 
 
+def _find_orphans(hist_rev: HistoricalResolver):
+    """
+    Within a revision, find new refs that can't be reached from refs in the
+    older ones.
+    """
+
+    # Note: this function assumes that there is no shady behaviour with older
+    #  revisions referring to as-of-yet-undefined references in future
+    #  revisions.
+    # TODO I might want to put a failsafe in the PdfFileReader class's
+    #  dereferencing logic to prevent that.
+
+    # This assumption makes finding orphans relatively cheap: we only need to
+    # pull up the dependencies of the older objects that were overwritten
+    # in this exact revision, and we only have to recurse into branches that
+    # pass through new objects themselves.
+
+    new_refs = hist_rev.explicit_refs_in_revision()
+
+    previous = hist_rev.reader.get_historical_resolver(hist_rev.revision - 1)
+
+    # These are newly updated refs that already existed in older revisions.
+    #  We want to know which of the new refs are reachable from one of these.
+    updated_old_refs = set()
+    # The candidate orphans are all the others
+    candidate_orphans = set()
+    for ref in new_refs:
+        if previous.is_ref_available(ref):
+            # ref didn't exist in previous revision
+            candidate_orphans.add(ref)
+        else:
+            updated_old_refs.add(ref)
+
+    def _objs_to_check() -> Iterator[PdfObject]:
+        # check the trailer too!
+        yield hist_rev.trailer_view
+        for _ref in updated_old_refs:
+            # take care to return the historical value here
+            yield hist_rev(_ref)
+
+    obj_iter = _objs_to_check()
+    while candidate_orphans:
+        try:
+            obj = next(obj_iter)
+        except StopIteration:
+            break
+        candidate_orphans -= hist_rev.collect_dependencies(
+            obj, since_revision=hist_rev.revision
+        )
+    return candidate_orphans
+
+
 class StandardDiffPolicy(DiffPolicy):
     """
     Run a list of rules to analyse the differences between two revisions.
@@ -1843,14 +1933,21 @@ class StandardDiffPolicy(DiffPolicy):
             in which case allowing objects to be freed might be reasonable.
             That said, pyHanko takes the conservative default position to reject
             all object freeing instructions as suspect.
+    :param ignore_orphaned_objects:
+        Some PDF writers create objects that aren't used anywhere (tsk tsk).
+        Since those don't affect the "actual" document content, they can usually
+        be ignored. If ``True``, newly created orphaned objects will be
+        cleared at level :attr:`.ModificationLevel.LTA_UPDATES`.
+        Default is ``True``.
     """
 
     def __init__(self, global_rules: List[QualifiedWhitelistRule],
                  form_rule: Optional[FormUpdatingRule],
-                 reject_object_freeing=True):
+                 reject_object_freeing=True, ignore_orphaned_objects=True):
         self.global_rules = global_rules
         self.form_rule = form_rule
         self.reject_object_freeing = reject_object_freeing
+        self.ignore_orphaned_objects = ignore_orphaned_objects
 
     def apply(self, old: HistoricalResolver, new: HistoricalResolver,
               field_mdp_spec: Optional[FieldMDPSpec] = None,
@@ -1880,6 +1977,10 @@ class StandardDiffPolicy(DiffPolicy):
                 if usages:
                     yield _ref, (ModificationLevel.NONE, set(usages))
 
+        # orphaned objects are cleared at LTA update level
+        for _ref in _find_orphans(new):
+            explained[ModificationLevel.LTA_UPDATES].add(_ref)
+
         # This table records all the overridden refs that already existed
         # in the old revision, together with the different ways they can be
         # reached from the document trailer.
