@@ -2,18 +2,25 @@ import binascii
 import hashlib
 import logging
 import uuid
-from dataclasses import dataclass
+from dataclasses import dataclass, field
 from datetime import datetime
 from io import BytesIO, BufferedIOBase
-from typing import Optional
+from typing import Optional, Set
 
 import tzlocal
-from asn1crypto import x509, cms, core, algos, pem, keys, pdf as asn1_pdf
+from asn1crypto import x509, cms, core, algos, keys, pdf as asn1_pdf
 from asn1crypto.algos import SignedDigestAlgorithm
-from certvalidator.errors import PathValidationError
+from cryptography.hazmat.primitives import serialization, hashes
+from cryptography.hazmat.primitives.asymmetric.ec import \
+    EllipticCurvePrivateKey, ECDSA
+from cryptography.hazmat.primitives.asymmetric.padding import PKCS1v15
+from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey
+from cryptography.hazmat.primitives.serialization import pkcs12
+
+from certvalidator.errors import PathValidationError, PathBuildingError
 
 from certvalidator import ValidationContext, CertificateValidator
-from oscrypto import asymmetric, keys as oskeys
+from pyhanko.sign.ades.api import CAdESSignedAttrSpec
 
 from pyhanko.pdf_utils import generic, misc
 from pyhanko.pdf_utils.generic import pdf_name, pdf_date, pdf_string
@@ -30,7 +37,11 @@ from pyhanko.sign.fields import (
 from pyhanko.sign.timestamps import TimeStamper
 from pyhanko.sign.general import (
     simple_cms_attribute, CertificateStore,
-    SimpleCertificateStore, SigningError, _sign_pss_raw, optimal_pss_params,
+    SimpleCertificateStore, SigningError, optimal_pss_params,
+    load_certs_from_pemder, load_cert_from_pemder,
+    _process_pss_params, load_private_key_from_pemder,
+    _translate_pyca_cryptography_key_to_asn1,
+    _translate_pyca_cryptography_cert_to_asn1,
 )
 from pyhanko.stamp import (
     TextStampStyle, TextStamp, STAMP_ART_CONTENT,
@@ -150,23 +161,7 @@ class PdfByteRangeDigest(generic.DictionaryObject):
             output = writer.prev.stream
             writer.write_in_place()
         else:
-            # Render the PDF to a byte buffer with placeholder values
-            # for the signature data, or straight to the provided output stream
-            # if possible
-            if output is None:
-                output = BytesIO()
-            else:
-                # Rationale for the explicit writability check:
-                #  If the output buffer is not readable or not seekable, it's
-                #  about to be replaced with a BytesIO instance, and in that
-                #  case, the write error would only happen *after* the signing
-                #  operations are done. We want to avoid that scenario.
-                if not output.writable():
-                    raise IOError(
-                        "Output buffer is not writable"
-                    )  # pragma: nocover
-                if not output.seekable() or not output.readable():
-                    output = BytesIO()
+            output = misc.prepare_rw_output_stream(output)
 
             writer.write(output)
 
@@ -397,10 +392,7 @@ class Signer:
         # Grab the certificate's algorithm (but forget about the digest)
         #  and use that to set up the default.
         # We'll specify the digest somewhere else.
-        cert_pubkey: asymmetric.PublicKey = asymmetric.load_public_key(
-            self.signing_cert.public_key
-        )
-        algo = cert_pubkey.algorithm
+        algo = self.signing_cert.public_key.algorithm
         params = None
         if algo == 'ec':
             mech = 'ecdsa'
@@ -489,10 +481,14 @@ class Signer:
 
     def signed_attrs(self, data_digest: bytes, digest_algorithm: str,
                      timestamp: datetime = None,
-                     revocation_info=None, use_pades=False):
+                     revocation_info=None, use_pades=False,
+                     cades_meta: CAdESSignedAttrSpec=None,
+                     timestamper=None, dry_run=False):
         """
         .. versionchanged:: 0.4.0
             Added positional ``digest_algorithm`` parameter _(breaking change)_.
+        .. versionchanged:: 0.5.0
+            Added ``dry_run``, ``timestamper`` and ``cades_meta`` parameters.
 
         Format the signed attributes for a CMS signature.
 
@@ -510,6 +506,20 @@ class Signer:
             (ignored when ``use_pades`` is ``True``).
         :param use_pades:
             Respect PAdES requirements.
+        :param dry_run:
+            .. versionadded:: 0.5.0
+
+            Flag indicating "dry run" mode. If ``True``, only the approximate
+            size of the output matters, so cryptographic
+            operations can be replaced by placeholders.
+        :param timestamper:
+            .. versionadded:: 0.5.0
+
+            Timestamper to use when creating timestamp tokens.
+        :param cades_meta:
+            .. versionadded:: 0.5.0
+
+            Specification for CAdES-specific attributes.
         :return:
             An :class:`.asn1crypto.cms.CMSAttributes` object.
         """
@@ -556,6 +566,14 @@ class Signer:
                 )
             )
 
+        # apply CAdES-specific attributes regardless of use_pades
+        if cades_meta is not None:
+            cades_attrs = cades_meta.extra_signed_attributes(
+                data_digest, digest_algorithm, timestamper=timestamper,
+                dry_run=dry_run
+            )
+            attrs.extend(cades_attrs)
+
         return cms.CMSAttributes(attrs)
 
     def unsigned_attrs(self, digest_algorithm, signature: bytes,
@@ -639,8 +657,9 @@ class Signer:
 
     def sign(self, data_digest: bytes, digest_algorithm: str,
              timestamp: datetime = None, dry_run=False,
-             revocation_info=None, use_pades=False,
-             timestamper=None) -> cms.ContentInfo:
+             revocation_info=None, use_pades=False, timestamper=None,
+             cades_signed_attr_meta: CAdESSignedAttrSpec = None) \
+            -> cms.ContentInfo:
 
         """
         Produce a detached CMS signature from a raw data digest.
@@ -677,6 +696,10 @@ class Signer:
                 this might still hit the timestamping server (in order to
                 produce a realistic size estimate), but the dummy response will
                 be cached.
+        :param cades_signed_attr_meta:
+            .. versionadded:: 0.5.0
+
+            Specification for CAdES-specific attributes.
         :return:
             An :class:`~.asn1crypto.cms.ContentInfo` object.
         """
@@ -686,7 +709,9 @@ class Signer:
         # signed attributes of our message
         signed_attrs = self.signed_attrs(
             data_digest, digest_algorithm, timestamp,
-            revocation_info=revocation_info, use_pades=use_pades
+            revocation_info=revocation_info, use_pades=use_pades,
+            timestamper=timestamper, cades_meta=cades_signed_attr_meta,
+            dry_run=dry_run
         )
 
         digest_algorithm_obj = algos.DigestAlgorithm(
@@ -748,6 +773,11 @@ Default message digest algorithm used when computing digests for use in
 signatures.
 """
 
+DEFAULT_SIGNER_KEY_USAGE = {"non_repudiation"}
+"""
+Default key usage bits required for the signer's certificate.
+"""
+
 
 @dataclass(frozen=True)
 class PdfSignatureMetadata:
@@ -871,35 +901,24 @@ class PdfSignatureMetadata:
         on approval signatures.
     """
 
-
-def load_certs_from_pemder(cert_files):
+    signer_key_usage: Set[str] = field(
+        default_factory=lambda: DEFAULT_SIGNER_KEY_USAGE
+    )
     """
-    A convenience function to load PEM/DER-encoded certificates from files.
+    Key usage extensions required for the signer's certificate.
+    Defaults to ``non_repudiation`` only, but sometimes ``digital_signature``
+    or a combination of both may be more appropriate.
+    See :class:`x509.KeyUsage` for a complete list.
+    
+    Only relevant if a validation context is also provided.
+    """
 
-    :param cert_files:
-        An iterable of file names.
-    :return:
-        A generator producing :class:`.asn1crypto.x509.Certificate` objects.
+    cades_signed_attr_spec: Optional[CAdESSignedAttrSpec] = None
+    """
+    .. versionadded:: 0.5.0
+
+    Specification for CAdES-specific attributes.
     """
-    for ca_chain_file in cert_files:
-        with open(ca_chain_file, 'rb') as f:
-            ca_chain_bytes = f.read()
-        # use the pattern from the asn1crypto docs
-        # to distinguish PEM/DER and read multiple certs
-        # from one PEM file (if necessary)
-        if pem.detect(ca_chain_bytes):
-            pems = pem.unarmor(ca_chain_bytes, multiple=True)
-            for type_name, _, der in pems:
-                if type_name is None or type_name.lower() == 'certificate':
-                    yield x509.Certificate.load(der)
-                else:  # pragma: nocover
-                    logger.debug(
-                        f'Skipping PEM block of type {type_name} in '
-                        f'{ca_chain_file}.'
-                    )
-        else:
-            # no need to unarmor, just try to load it immediately
-            yield x509.Certificate.load(ca_chain_bytes)
 
 
 class SimpleSigner(Signer):
@@ -929,26 +948,32 @@ class SimpleSigner(Signer):
 
         signature_mechanism = self.get_signature_mechanism(digest_algorithm)
         mechanism = signature_mechanism.signature_algo
+        priv_key = serialization.load_der_private_key(
+            self.signing_key.dump(), password=None
+        )
+
         if mechanism == 'rsassa_pkcs1v15':
-            sign_func = asymmetric.rsa_pkcs1v15_sign
+            padding = PKCS1v15()
+            hash_algo = getattr(hashes, digest_algorithm.upper())()
+            assert isinstance(priv_key, RSAPrivateKey)
+            return priv_key.sign(data, padding, hash_algo)
         elif mechanism == 'rsassa_pss':
-            return _sign_pss_raw(
-                data, self.signing_key, signature_mechanism['parameters'],
-                digest_algorithm
+            params = signature_mechanism['parameters']
+            padding, hash_algo = _process_pss_params(
+                params, digest_algorithm
             )
+            assert isinstance(priv_key, RSAPrivateKey)
+            return priv_key.sign(data, padding, hash_algo)
         elif mechanism == 'ecdsa':
-            sign_func = asymmetric.ecdsa_sign
+            hash_algo = getattr(hashes, digest_algorithm.upper())()
+            assert isinstance(priv_key, EllipticCurvePrivateKey)
+            return priv_key.sign(data, signature_algorithm=ECDSA(hash_algo))
         else:  # pragma: nocover
             raise SigningError(
                 f"The signature mechanism {mechanism} "
                 "is unsupported by this signer."
             )
 
-        return sign_func(
-            asymmetric.load_private_key(self.signing_key),
-            data, digest_algorithm.lower()
-        )
-
     @classmethod
     def _load_ca_chain(cls, ca_chain_files=None):
         try:
@@ -994,7 +1019,15 @@ class SimpleSigner(Signer):
         if ca_chain is None:  # pragma: nocover
             return None
 
-        (kinfo, cert, other_certs) = oskeys.parse_pkcs12(pfx_bytes, passphrase)
+        (private_key, cert, other_certs) = pkcs12.load_key_and_certificates(
+            pfx_bytes, passphrase
+        )
+        kinfo = _translate_pyca_cryptography_key_to_asn1(private_key)
+        cert = _translate_pyca_cryptography_cert_to_asn1(cert)
+        other_certs = set(
+            map(_translate_pyca_cryptography_cert_to_asn1, other_certs)
+        )
+
         cs = SimpleCertificateStore()
         cs.register_multiple(ca_chain | set(other_certs))
         return SimpleSigner(
@@ -1032,15 +1065,11 @@ class SimpleSigner(Signer):
         """
         try:
             # load cryptographic data (both PEM and DER are supported)
-            with open(key_file, 'rb') as f:
-                signing_key: keys.PrivateKeyInfo = oskeys.parse_private(
-                    f.read(), password=key_passphrase
-                )
-            with open(cert_file, 'rb') as f:
-                signing_cert: x509.Certificate = oskeys.parse_certificate(
-                    f.read()
-                )
-        except (IOError, ValueError) as e:  # pragma: nocover
+            signing_key = load_private_key_from_pemder(
+                key_file, passphrase=key_passphrase
+            )
+            signing_cert = load_cert_from_pemder(cert_file)
+        except (IOError, ValueError, TypeError) as e:  # pragma: nocover
             logger.error('Could not load cryptographic material', exc_info=e)
             return None
 
@@ -1195,7 +1224,8 @@ def _get_or_create_sigfield(field_name, pdf_out: BasePdfFileWriter,
                 'box': new_field_spec.box,
                 'include_on_page': pdf_out.find_page_for_modification(
                     new_field_spec.on_page
-                )[0]
+                )[0],
+                'combine_annotation': new_field_spec.combine_annotation
             }
         else:
             sig_field_kwargs = {}
@@ -1308,12 +1338,15 @@ class SigAppearanceSetup:
     stamp style.
     """
 
-    def _apply(self, sig_field, writer):
+    def _apply(self, sig_annot, writer):
         style = self.style
         name = self.name
         timestamp = self.timestamp
         extra_text_params = self.text_params or {}
-        x1, y1, x2, y2 = sig_field[pdf_name('/Rect')]
+        try:
+            x1, y1, x2, y2 = sig_annot['/Rect']
+        except KeyError:
+            return
         w = abs(x1 - x2)
         h = abs(y1 - y2)
         if w and h:
@@ -1343,11 +1376,10 @@ class SigAppearanceSetup:
                     writer, style=style, text_params=text_params,
                     box=box
                 )
-            sig_field[
-                pdf_name('/AP')] = stamp.as_appearances().as_pdf_object()
+            sig_annot['/AP'] = stamp.as_appearances().as_pdf_object()
             try:
                 # if there was an entry like this, it's meaningless now
-                del sig_field[pdf_name('/AS')]
+                del sig_annot[pdf_name('/AS')]
             except KeyError:
                 pass
 
@@ -1515,7 +1547,19 @@ class PdfCMSEmbedder:
         # take care of the field's visual appearance (if applicable)
         appearance_setup = sig_obj_setup.appearance_setup
         if appearance_setup is not None:
-            appearance_setup._apply(sig_field, writer)
+            try:
+                sig_annot, = sig_field['/Kids']
+                sig_annot = sig_annot.get_object()
+            except (ValueError, TypeError):
+                raise SigningError(
+                    "Failed to access signature field's annotation. "
+                    "Signature field must have exactly one child annotation, "
+                    "or it must be combined with its annotation."
+                )
+            except KeyError:
+                sig_annot = sig_field
+
+            appearance_setup._apply(sig_annot, writer)
 
         sig_obj = sig_obj_setup.sig_placeholder
         sig_obj_ref = writer.add_object(sig_obj)
@@ -1676,7 +1720,9 @@ class PdfTimeStamper:
         return output
 
     def update_archival_timestamp_chain(self, reader: PdfFileReader,
-                                        validation_context, in_place=True):
+                                        validation_context, in_place=True,
+                                        output=None, chunk_size=4096,
+                                        default_md_algorithm=DEFAULT_MD):
         """
         Validate the last timestamp in the timestamp chain on a PDF file, and
         write an updated version to an output stream.
@@ -1686,9 +1732,20 @@ class PdfTimeStamper:
         :param validation_context:
             :class:`.certvalidator.ValidationContext` object to validate
             the last timestamp.
+        :param output:
+            Write the output to the specified output stream.
+            If ``None``, write to a new :class:`.BytesIO` object.
+            Default is ``None``.
         :param in_place:
-            Sign the input in-place. If ``False``, write output to a
-            :class:`.BytesIO` object.
+            Sign the original input stream in-place.
+            This parameter overrides ``output``.
+        :param chunk_size:
+            Size of the internal buffer (in bytes) used to feed data to the
+            message digest function if the input stream does not support
+            ``memoryview``. Default is 4096.
+        :param default_md_algorithm:
+            Message digest to use if there are no preceding timestamps in the
+            file.
         :return:
             The output stream containing the signed output.
         """
@@ -1697,51 +1754,59 @@ class PdfTimeStamper:
         # TODO: add an option to validate the entire timestamp chain
         #  plus all signatures
         from .validation import (
-            EmbeddedPdfSignature, _establish_timestamp_trust,
-            DocumentSecurityStore
+            _establish_timestamp_trust, DocumentSecurityStore,
+            get_timestamp_chain
         )
 
-        all_signatures = reader.embedded_signatures
-        if not all_signatures:
-            raise SigningError("No signatures found.")
-        last_signature: EmbeddedPdfSignature = \
-            all_signatures[len(all_signatures) - 1]
-        last_signature.compute_digest()
-        last_signature.compute_tst_digest()
+        timestamps = get_timestamp_chain(reader)
+        try:
+            last_timestamp = next(timestamps)
+        except StopIteration:
+            logger.warning(
+                "Document does not have any document timestamps yet. "
+                "This may cause unexpected results."
+            )
+            last_timestamp = None
 
-        # two cases: either the signature is a normal signature,
-        # or it is a document timestamp
-        if last_signature.sig_object['/Type'] == '/DocTimeStamp':
-            tst_token = last_signature.signed_data
-            expected_imprint = last_signature.raw_digest
+        # Validate the previous timestamp if present
+        tst_status = None
+        if last_timestamp is None:
+            md_algorithm = default_md_algorithm
         else:
-            # normal signature
-            tst_token = last_signature.external_timestamp_data
-            if tst_token is None:  # pragma: nocover
-                raise SigningError("Final signature does not have a timestamp.")
-            expected_imprint = last_signature.tst_signature_digest
+            last_timestamp.compute_digest()
+            last_timestamp.compute_tst_digest()
 
-        # run validation logic
-        tst_status = _establish_timestamp_trust(
-            tst_token, validation_context, expected_imprint
-        )
+            tst_token = last_timestamp.signed_data
+            expected_imprint = last_timestamp.external_digest
 
+            # run validation logic
+            tst_status = _establish_timestamp_trust(
+                tst_token, validation_context, expected_imprint
+            )
+
+            md_algorithm = tst_status.md_algorithm
+
+        # Prepare output
         if in_place:
             output = reader.stream
-        else:  # pragma: nocover
-            output = BytesIO()
-            output.write(reader.stream.read())
+        else:
+            output = misc.prepare_rw_output_stream(output)
+            reader.stream.seek(0)
+            misc.chunked_write(
+                bytearray(chunk_size), reader.stream, output
+            )
 
-        # update the DSS
-        DocumentSecurityStore.add_dss(
-            output, last_signature.pkcs7_content,
-            paths=(tst_status.validation_path,),
-            validation_context=validation_context
-        )
+        if last_timestamp is not None:
+            # update the DSS
+            DocumentSecurityStore.add_dss(
+                output, last_timestamp.pkcs7_content,
+                paths=(tst_status.validation_path,),
+                validation_context=validation_context
+            )
 
         # append a new timestamp
         return self.timestamp_pdf(
-            IncrementalPdfFileWriter(output), tst_status.md_algorithm,
+            IncrementalPdfFileWriter(output), md_algorithm,
             validation_context, in_place=True
         )
 
@@ -2063,12 +2128,19 @@ class PdfSigner(PdfTimeStamper):
         signature_meta: PdfSignatureMetadata = self.signature_meta
         signer: Signer = self.signer
         validation_context = signature_meta.validation_context
-        if signature_meta.embed_validation_info and validation_context is None:
-            raise SigningError(
-                'A validation context must be provided if '
-                'validation/revocation info is to be embedded into the '
-                'signature.'
-            )
+        if signature_meta.embed_validation_info:
+            if validation_context is None:
+                raise SigningError(
+                    'A validation context must be provided if '
+                    'validation/revocation info is to be embedded into the '
+                    'signature.'
+                )
+            elif not validation_context._allow_fetching:
+                logger.warning(
+                    "Validation/revocation info will be embedded, but "
+                    "fetching is not allowed. This may give rise to unexpected "
+                    "results."
+                )
         validation_paths = []
         signer_cert_validation_path = None
         if validation_context is not None:
@@ -2078,15 +2150,49 @@ class PdfSigner(PdfTimeStamper):
                 signer.signing_cert, intermediate_certs=signer.cert_registry,
                 validation_context=validation_context
             )
-            # TODO allow customisation of key usage parameters
             try:
                 signer_cert_validation_path = validator.validate_usage(
-                    {"non_repudiation"}
+                    signature_meta.signer_key_usage
+                )
+            except (PathBuildingError, PathValidationError) as e:
+                raise SigningError(
+                    "The signer's certificate could not be validated", e
                 )
-            except PathValidationError as e:
-                raise SigningError("The signer's certificate was not valid", e)
             validation_paths.append(signer_cert_validation_path)
 
+            # If LTA:
+            # if the original document already included a document timestamp,
+            # we need to collect revocation information for it, to preserve
+            # the integrity of the timestamp chain
+            from .validation import get_timestamp_chain
+            if signature_meta.use_pades_lta \
+                    and isinstance(pdf_out, IncrementalPdfFileWriter):
+
+                # try to grab the most recent document timestamp
+                last_ts = None
+                try:
+                    last_ts = next(get_timestamp_chain(pdf_out.prev))
+                except StopIteration:
+                    pass
+
+                if last_ts is not None:
+                    ts_validator = CertificateValidator(
+                        last_ts.signer_cert,
+                        intermediate_certs=signer.cert_registry,
+                        validation_context=validation_context
+                    )
+                    try:
+                        last_ts_validation_path = ts_validator.validate_usage(
+                            set(), extended_key_usage={"time_stamping"}
+                        )
+                    except (PathBuildingError, PathValidationError) as e:
+                        raise SigningError(
+                            "Requested a PAdES-LTA signature on an existing "
+                            "document, but the most recent timestamp "
+                            "could not be validated.", e
+                        )
+                    validation_paths.append(last_ts_validation_path)
+
         cms_writer = PdfCMSEmbedder(
             new_field_spec=self.new_field_spec
         ).write_cms(
@@ -2172,7 +2278,8 @@ class PdfSigner(PdfTimeStamper):
                 test_md, md_algorithm,
                 timestamp=timestamp, use_pades=use_pades,
                 dry_run=True, revocation_info=revinfo,
-                timestamper=timestamper
+                timestamper=timestamper,
+                cades_signed_attr_meta=signature_meta.cades_signed_attr_spec
             )
             test_len = len(test_signature_cms.dump()) * 2
             # External actors such as timestamping servers can't be relied on to
@@ -2214,7 +2321,8 @@ class PdfSigner(PdfTimeStamper):
         signature_cms = signer.sign(
             true_digest, md_algorithm,
             timestamp=timestamp, use_pades=use_pades,
-            revocation_info=revinfo, timestamper=timestamper
+            revocation_info=revinfo, timestamper=timestamper,
+            cades_signed_attr_meta=signature_meta.cades_signed_attr_spec
         )
         # ... and feed it to the CMS writer
         res_output, sig_contents = cms_writer.send(signature_cms)
